/***********************************************************************
 *
 * Name:        Scheduler.c
 *
 * Description: Time controlling module for Video Plugin               
 *
 * Author:	Tomasz Stachowiak, NPAC
 *
 * Date:	November 1996
 *
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/RowColumn.h>
#include <Xm/Text.h>
#include <Xm/PushB.h>
#include <X11/Shell.h>

#include "VideoProtocol.h"

#define MAX_USERS 20

#define DELAY_MAX 200
#define DELAY_MIN 50
#define RATE_STEP 1
#define SPACE_MIN 5
#define INTERVAL_STEP 50
#define SPEED_RATIO 0.1

typedef struct timeval statTime;

typedef struct _usrTimeInfo
{
  int userNo;                    /* conference member ID */
  int rate;
  statTime startT;                   /* time of receiving first frame */ 
  statTime startStamp;               /* first frame time stamp */
  statTime lastStamp;                /* previous frame time stamp */
  statTime lastFrameT;               /* time of receiving first frame */
  int framesReceived;
  statTime totalT;
  int framesLate;
  int framesAvanced;
  int jitter;
  int totalBytes;
  int lastFrameDiff;
  int lastDecodeTime;
} UsrTimeInfo;

typedef struct _sendTimeInfo
{
  float rate;
  int interval;
  statTime startT;
  statTime lastFrameT;
  statTime nextFrameT;
  int framesSent;
  statTime totalT;
  int slowNo;
  int speedNo;
  int totalBytes;
  int lastFrameDiff;
} SendTimeInfo;

void slowDown();
void speedUp();
void timeSum(statTime *a, int b, statTime *result);
int timeDiff(statTime *a, statTime *b);

void statClbkOK(Widget w, XtPointer app_data, XtPointer widget_data);
void statClbkUpdate(Widget w, XtPointer app_data, XtPointer widget_data);
void statClbkReset(Widget w, XtPointer app_data, XtPointer widget_data);

UsrTimeInfo usrTimeInfo[MAX_USERS+1];
SendTimeInfo sendTimeInfo;
static int rec_first[MAX_USERS+1];
static int send_first = 1;
int max_rate = 0;
int canChange = 1;

extern float delayCoeff;
extern state;

/*********************************************************************
 * Procedure handles received frame, it does all the statistics and
 * optionally it changes the sending rate 
 *
 *********************************************************************/
void doReceiveStatistics(int userNo, statTime *frameTimeStamp, int decode_time)
{
  statTime currentT;
  int diff, local_diff, remote_diff;

  gettimeofday(&currentT);
  if(!rec_first[userNo])
    {
      local_diff = timeDiff(&currentT, &usrTimeInfo[userNo].lastFrameT);
      usrTimeInfo[userNo].lastFrameDiff = local_diff;
    }

  if(rec_first[userNo] == 1)
    {
      gettimeofday(&usrTimeInfo[userNo].startT);
      local_diff = timeDiff(&currentT, &usrTimeInfo[userNo].lastFrameT);
      usrTimeInfo[userNo].lastFrameDiff = local_diff;
      rec_first[userNo]--;
    }
  if(rec_first[userNo] > 1)
    {
      rec_first[userNo]--;
    }

  usrTimeInfo[userNo].framesReceived++;
  usrTimeInfo[userNo].lastFrameT.tv_sec = currentT.tv_sec;
  usrTimeInfo[userNo].lastFrameT.tv_usec = currentT.tv_usec;

  usrTimeInfo[userNo].lastDecodeTime = decode_time;
  /* ...calculate jitter */
}

void receiveStart(int userNo)
{
}

void receiveStop(int userNo)
{
  rec_first[userNo] = 2;
}

void receiveTimeInit()
{
  int i;
  
  for(i=0; i<MAX_USERS+1; i++)
    {
      usrTimeInfo[i].userNo = -1;
      rec_first[i] = 2;
    }
}

void usrTimeInit(int userNo)
{
  usrTimeInfo[userNo].userNo = userNo;
  gettimeofday(&usrTimeInfo[userNo].startT);
  usrTimeInfo[userNo].framesReceived = 0;
  usrTimeInfo[userNo].totalT.tv_sec = 0;
  usrTimeInfo[userNo].totalT.tv_usec = 0;
  usrTimeInfo[userNo].framesLate = 0;
  usrTimeInfo[userNo].framesAvanced = 0;
  usrTimeInfo[userNo].jitter = 0;
  usrTimeInfo[userNo].totalBytes = 0;
  
}

void usrTimeReport()
{
  int i;

  for(i=0; i<MAX_USERS+1; i++)
    if(usrTimeInfo[i].userNo >= 0)
      {
	printf("----------------------------------------------\n");
	printf("USER: %d\n", i);
	printf("----------------------------------------------\n");
	printf("Frames received: %d\n",usrTimeInfo[i].framesReceived);
	printf("Total time:      %d\n", 
	       (usrTimeInfo[i].totalT.tv_sec*1000)
	       +(usrTimeInfo[i].totalT.tv_usec/1000));
	printf("Frames late:     %d\n", usrTimeInfo[i].framesLate);
	printf("Frames avanced:  %d\n", usrTimeInfo[i].framesAvanced);
	printf("----------------------------------------------\n\n");
	fflush(stdout);
      }
}


int statCalcInterval()
{
  int i;
  int interval=0;

  for(i=0; i<MAX_USERS+1; i++)
    if(rec_first[i] == 0)
      {
	interval += (int)(((float)sendTimeInfo.lastFrameDiff/(float)usrTimeInfo[i].lastFrameDiff) * (float)(usrTimeInfo[i].lastDecodeTime));
	/*
	printf("Interval part: %d\n", interval);
	printf("Coded diff: %d\n", sendTimeInfo.lastFrameDiff);
	printf("Decoded diff: %d\n", usrTimeInfo[i].lastFrameDiff);
	printf("Decoded time: %d\n", usrTimeInfo[i].lastDecodeTime);
	*/
      }
  interval *= delayCoeff;

  if(interval == 0)
    interval = 5;

  /*printf("Interval: %d\n", interval);*/
  return interval;    
}

void doSendStatistics(int bytes)
{
  statTime currentT;

  gettimeofday(&currentT);
  if(!send_first)
    sendTimeInfo.lastFrameDiff = timeDiff(&currentT, &sendTimeInfo.lastFrameT);
  sendTimeInfo.rate = (float)1000 / (float)sendTimeInfo.lastFrameDiff;
  /*  printf("Sending rate: %f\n", sendTimeInfo.rate); */
  sendTimeInfo.lastFrameT.tv_sec = currentT.tv_sec;
  sendTimeInfo.lastFrameT.tv_usec = currentT.tv_usec;
  sendTimeInfo.framesSent++;
  sendTimeInfo.totalBytes += bytes;
  send_first = 0;
}


void sendStart()
{
  gettimeofday(&sendTimeInfo.startT);
  sendTimeInfo.interval = 200;
}

void sendStop()
{
  statTime tmp;

  gettimeofday(&tmp);
  timeSum(&sendTimeInfo.totalT, 
	  timeDiff(&tmp, &sendTimeInfo.startT), 
	  &sendTimeInfo.totalT);
  sendTimeInfo.lastFrameDiff = 0;  
  send_first = 1;
}

void sendTimeInit()
{
  sendTimeInfo.lastFrameDiff = 0;  
  sendTimeInfo.rate = 5;
  sendTimeInfo.interval = 200;
  gettimeofday(&sendTimeInfo.startT);
  sendTimeInfo.framesSent = 0;
  sendTimeInfo.totalT.tv_sec = 0;
  sendTimeInfo.totalT.tv_usec = 0;
  sendTimeInfo.slowNo = 0;
  sendTimeInfo.speedNo = 0;
  sendTimeInfo.totalBytes = 0;
  
}

void sendTimeReport()
{
  char msg[500];
  float frameRate, bytesRate, bytesFrame;

  frameRate = (float)sendTimeInfo.framesSent/(float)sendTimeInfo.totalT.tv_sec;
  bytesRate = (float)sendTimeInfo.totalBytes/(float)sendTimeInfo.totalT.tv_sec;
  bytesFrame = (float)sendTimeInfo.totalBytes/(float)sendTimeInfo.framesSent;
  sprintf(msg, "        STATISTICS\n--------------------------\nFrames sent:   %d\nTotal time:    %d ms\nBytes/frame:   %f\nBytes/second:  %f\nFrames/second: %f\n", 
	  sendTimeInfo.framesSent, 
	  (sendTimeInfo.totalT.tv_sec*1000)+(sendTimeInfo.totalT.tv_usec/1000),
	  bytesFrame, 
	  bytesRate, 
	  frameRate);
  showMessage(msg);
}
char *statGetInfo()
{
  char *msg;
  float frameRate, bytesRate, bytesFrame;

  msg = malloc(500);
  frameRate = (float)sendTimeInfo.framesSent/(float)sendTimeInfo.totalT.tv_sec;
  bytesRate = (float)sendTimeInfo.totalBytes/(float)sendTimeInfo.totalT.tv_sec;
  bytesFrame = (float)sendTimeInfo.totalBytes/(float)sendTimeInfo.framesSent;
  if(sendTimeInfo.totalT.tv_sec == 0)
    {
      frameRate = 0; 
      bytesRate = 0;
    }
  if(sendTimeInfo.framesSent == 0)
    {
      bytesFrame = 0;
    }

  sprintf(msg, "        STATISTICS\n__________________________\nFrames sent:   %d\nTotal time:    %d ms\nBytes/frame:   %f\nBytes/second:  %f\nFrames/second: %f\n__________________________\n", 
	  sendTimeInfo.framesSent, 
	  (sendTimeInfo.totalT.tv_sec*1000)+(sendTimeInfo.totalT.tv_usec/1000),
	  bytesFrame, 
	  bytesRate, 
	  frameRate);
  
  return msg;
}

void statSendReset()
{
  sendTimeInfo.framesSent = 0;
  sendTimeInfo.totalBytes = 0;
  sendTimeInfo.totalT.tv_sec = 0;
  sendTimeInfo.totalT.tv_usec = 0;
}

/********************************************************************
 * 
 * Procedure operating on timeval structures
 *
 *******************************************************************/
/* Sum result = a + b (b in miliseconds)*/
void timeSum(statTime *a, int b, statTime *result)
{
  int pom;

  result->tv_sec = a->tv_sec + (int)(b / 1000);
  result->tv_usec = (b % 1000) * 1000;
}

/* Difference a-b in miliseconds */
int timeDiff(statTime *a, statTime *b)
{
  int result;

  result = 1000 * (a->tv_sec - b->tv_sec) + (a->tv_usec - b->tv_usec)/1000;

  return result;
}



/********************************************************************
 * 
 * Statistics dialog window
 *
 *******************************************************************/
void statShowDlg(Widget parent)
{
  Widget dlg, bUpdate, bReset, bOK, rect, text;
  XmString label;
  char *info;

  dlg = XtVaCreatePopupShell( "stat",
				 xmDialogShellWidgetClass,
				 parent,
				 XmNx, 370,
				 XmNy, 300,
				 XmNwidth, 300,
				 XmNheight, 200,
				 XmNtitle, "Statistics",
				 NULL );

  rect = XtVaCreateWidget("rect", xmRowColumnWidgetClass,
				       dlg, NULL);


  info = statGetInfo();
  text =  XtVaCreateManagedWidget("text", xmTextWidgetClass,
			   rect, 
			   XmNeditMode, XmMULTI_LINE_EDIT,
			   XmNeditable, False,
			   XmNrows , 8,
			   XmNcolumns , 26,
			   XmNautoShowCursorPosition, False,
			   XmNvalue, info,
			   NULL);
  free(info);
  

  label = XmStringCreateSimple("Update");
  bUpdate = XtVaCreateManagedWidget("update", xmPushButtonWidgetClass,
				       rect,
				       XmNlabelString, label, 
				       XmNx, 10, 
				       XmNy, 160,
				       XmNwidth, 50,
				       XmNheight, 30,
				       NULL);
  
  XmStringFree(label);
  XtAddCallback(bUpdate, XmNactivateCallback, statClbkUpdate, text);

  label = XmStringCreateSimple("Reset");
  bReset = XtVaCreateManagedWidget("reset", xmPushButtonWidgetClass,
				       rect,
				       XmNlabelString, label, 
				       XmNx, 70, 
				       XmNy, 160,
				       XmNwidth, 50,
				       XmNheight, 30,
				       NULL);
  
  XmStringFree(label);
  XtAddCallback(bReset, XmNactivateCallback, statClbkReset, text);

  label = XmStringCreateSimple("OK");
  bOK = XtVaCreateManagedWidget("ok", xmPushButtonWidgetClass,
				       rect,
				       XmNlabelString, label, 
				       XmNx, 130, 
				       XmNy, 160,
				       XmNwidth, 50,
				       XmNheight, 30,
				       NULL);
  
  XmStringFree(label);
  XtAddCallback(bOK, XmNactivateCallback, statClbkOK, NULL);

  XtManageChild(rect);

  XtManageChild(dlg);

  XtRealizeWidget(dlg);

  XtPopup(dlg, XtGrabNone);

}

/* OK callback */
void statClbkOK(Widget w, XtPointer app_data, XtPointer widget_data)
{
  XtDestroyWidget(XtParent(XtParent(w)));
}

/* Update callback */
void statClbkUpdate(Widget w, XtPointer text, XtPointer widget_data)
{
  char *info;

  if(state == PLAY)
    {
      sendStop();
      info = statGetInfo();
      sendStart();
    }
  else
      info = statGetInfo();

  XmTextSetString((Widget)text, info);
  free(info);
}

/* Reset callback */
void statClbkReset(Widget w, XtPointer text, XtPointer widget_data)
{
  char *info;

  statSendReset();
  info = statGetInfo();
  XmTextSetString((Widget)text, info);
  free(info);
}













